-- data width converter for master dwidth > slave dwidth
-- author: Ernest Jamro jamro@agh.edu.pl

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////

  
-- direct connection for read mdwidth>sdwidth
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity dwidth_conv_m_rd_direct is
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    mdi: in std_logic_vector(0 to mdwidth-1); -- data input
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end dwidth_conv_m_rd_direct;

architecture arch of dwidth_conv_m_rd_direct is
  signal buf_valid: std_logic; -- valid data in the read buffer
  signal buf: std_logic_vector(sdwidth to mdwidth-1); -- read buffer used only for the LSBs (the MSB is directly copied to the output)
begin
 gseqAddr0: if use_seqAddr=0  generate
      msel<= ssel;
      sack<= mack;
        
      gs8_m16: if sdwidth=8 and mdwidth=16 generate
          sdo<= mdi(0 to 7) when sa(awidth-1)='0' else mdi(8 to 15);
          mbe(0)<= not sa(awidth-1);  mbe(1)<= sa(awidth-1);
       end generate;
        
        gs16m32: if sdwidth=16 and mdwidth=32 generate
          sdo<= mdi(0 to 15) when sa(awidth-2)='0' else mdi(16 to 31);
          mbe(0 to 1)<= sbe when sa(awidth-2)='0' else "00";
          mbe(2 to 3)<= sbe when sa(awidth-2)='1' else "00";
        end generate;
        gs8m32: if sdwidth=8 and mdwidth=32 generate
          sdo<= mdi(0 to 7) when sa(awidth-2 to awidth-1)="00" else
               mdi(8 to 15) when sa(awidth-2 to awidth-1)="01" else
               mdi(16 to 23) when sa(awidth-2 to awidth-1)="10" else
               mdi(24 to 31);
          mbe(0)<= '1' when sa(awidth-2 to awidth-1)="00" else '0';
          mbe(1)<= '1' when sa(awidth-2 to awidth-1)="01" else '0';
          mbe(2)<= '1' when sa(awidth-2 to awidth-1)="10" else '0';
          mbe(3)<= '1' when sa(awidth-2 to awidth-1)="11" else '0';
        end generate;
   end generate;
----------------------------------------------------------------------------------------------------
  -- a single read on master for several reads on the slave. use internal read buffer (flip-flops to remember the all read data exept the MSBs
  gseqAddr1: if use_seqAddr>0 generate
     msel<= ssel and not buf_valid;
        
     gs8_m16: if sdwidth=8 and mdwidth=16 generate
        sack<= ssel and (buf_valid or -- the LSBs are read from the read buffer
            (mack and not sa(awidth-1))); -- the MSB is read directly from the master
        sdo<= mdi(0 to 7) when sa(awidth-1)='0' else buf(8 to 15);
        mbe(1)<= sa(awidth-1) or sseqAddr;  
        mbe(0)<= not sa(awidth-1);
        -- read buffer 
        process(arst, clk) begin
          if arst='1' then buf(8 to 15)<= (others=>'0');  buf_valid<= '0';
          elsif clk'event and clk='1' then
             if mack='1' and (sseqAddr='1' or sa(awidth-1)='1') then
               buf_valid<= '1';
               buf(8 to 15)<= mdi(8 to 15);
             elsif (sack='1' and sa(awidth-1)='1') or sseqAddr='0' then
               buf_valid<= '0';
             end if;
          end if;
        end process;
      end generate;
        
      gs16m32: if sdwidth=16 and mdwidth=32 generate
        sack<= ssel and (buf_valid or -- the LSBs are read from the read buffer
            (mack and not sa(awidth-2))); -- the MSB sis read directly from the master
        sdo<= mdi(0 to 15) when sa(awidth-2)='0' else buf(16 to 31);
        mbe(3)<= '1' when (sa(awidth-2)='1' and sbe(1)='1') or sseqAddr='1' else '0';
        mbe(2)<= '1' when (sa(awidth-2)='1' and sbe(0)='1') or (sseqAddr='1' and sa(awidth-2)='0') else '0';  
        mbe(1)<= not sa(awidth-2) and sbe(1);
        mbe(0)<= not sa(awidth-2) and sbe(0);
        -- read buffer 
        process(arst, clk) begin
          if arst='1' then buf(16 to 31)<= (others=>'0');  buf_valid<= '0';
          elsif clk'event and clk='1' then
             if mack='1' and (sseqAddr='1' or sa(awidth-2)='1') then
               buf_valid<= '1';
               buf(16 to 31)<= mdi(16 to 31);
             elsif (sack='1' and sa(awidth-2)='1') or sseqAddr='0' then
               buf_valid<= '0';
             end if;
          end if;
        end process;
      end generate;

      gs8m32: if sdwidth=8 and mdwidth=32 generate
         sack<= ssel and (buf_valid or -- the LSBs are read from the read buffer
            (mack and not sa(awidth-2) and not sa(awidth-1) )); -- the MSB is read directly from the master
          sdo<= buf(24 to 31) when sa(awidth-2 to awidth-1)="11" else 
                buf(16 to 23) when sa(awidth-2 to awidth-1)="10" else
                buf(8 to 15) when sa(awidth-2 to awidth-1)="01" else
                mdi(0 to 7);     
          mbe(3)<= '1' when sa(awidth-2 to awidth-1)="11" or sseqAddr='1' else '0';
          mbe(2)<= '1' when sa(awidth-2 to awidth-1)="10" or 
                   (sseqAddr='1' and sa(awidth-2 to awidth-1)/="11") else '0';
          mbe(1)<= '1' when sa(awidth-2 to awidth-1)="01" or 
                   (sseqAddr='1' and sa(awidth-2)='0') else '0';
          mbe(0)<= not sa(awidth-2) and not sa(awidth-1);
          -- read buffer 
          process(arst, clk) begin
           if arst='1' then buf(8 to 31)<= (others=>'0');  buf_valid<= '0';
           elsif clk'event and clk='1' then
             if mack='1' and (sseqAddr='1' or sa(awidth-2 to awidth-1)/="00") then
               buf_valid<= '1';
               buf(8 to 31)<= mdi(8 to 31);
             elsif (sack='1' and sa(awidth-2 to awidth-1)="11") or sseqAddr='0' then
               buf_valid<= '0';
             end if;
          end if;
        end process;
      end generate;

    end generate; -- if use_seqAddr>0 
 
end arch;


-------------------------------------------------------------------------------------
--  for read mdwidth>sdwidth with fifo
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity dwidth_conv_m_rd_fifo is
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    fifo_size: integer:= 4; -- address width of the fifo
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
    sa: in std_logic_vector(0 to awidth-1); -- address bus
	-- master interface (output)
    ma: in std_logic_vector(0 to awidth-1); -- address bus (input!!!)
    ma_valid: in std_logic; -- valid master address
    mdi: in std_logic_vector(0 to mdwidth-1); -- data input
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end dwidth_conv_m_rd_fifo;

architecture arch of dwidth_conv_m_rd_fifo is
component fifo
  generic( adr_width: integer:= 4; -- internal memory address width - 4 is preferable for distributed rams, 0- direct connection
	  data_width: integer:= 8); -- data bus width
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
	rst: in std_logic; -- synchronous reset for the whole device
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    half_full: out std_logic; -- fifo is half full
	-- data input signals (slave)
	datI_I: in std_logic_vector(0 to data_width-1);
	stbI_I: in std_logic; -- strobe similar as OPB_select (master ready to transfer data)
	ackI_O: out std_logic; -- acknowledge
	-- data output signals (master)
	datO_O: out std_logic_vector(0 to data_width-1);
	stbO_O: out std_logic;
	ackO_I: in std_logic);  
  end component;
  
  signal almost_full, full, empty, rst: std_logic;
  signal fifo_dout: std_logic_vector(0 to mdwidth-1); -- fifo data output
  signal fifo_stbO_O, fifo_ackO_I: std_logic;
  signal lsb_adr: std_logic; -- the last transfer (lsb is now transfered)
  signal sbe_q: std_logic_vector(0 to sdwidth/8 -1);
begin
  frd: fifo
    generic map(adr_width=> fifo_size, data_width=> mdwidth)
    port map(clk=> clk, arst=> arst, rst=> rst,
    almost_full=>almost_full, almost_empty=> open, full=> full, half_full=> open,
	-- data input signals (slave)
	datI_I=> mdi, stbI_I=> mack, ackI_O=> open,
	-- data output signals (master)
	datO_O=> fifo_dout, stbO_O=> fifo_stbO_O, ackO_I=> fifo_ackO_I);
    
  msel<= not full and ma_valid when use_seqAddr>0 else
    ma_valid and not fifo_stbO_O; -- read only a single data (fifo_stbO_O= fifo_not_empty)
  rst<= not ma_valid or (not sseqAddr and sack);
  
  sack<= ssel and ma_valid and fifo_stbO_O;
  fifo_ackO_I<= sack when use_seqAddr<=0 or sseqAddr='0' else sack and lsb_adr;

        
----------------------------------------------------------------------------------------------------
  gs16: if sdwidth=16 generate
    process(clk, arst) begin
      if arst='1' then sbe_q<= (others=>'0');
      elsif clk'event and clk='1' then
        sbe_q<= sbe;
      end if;
    end process;
  end generate;

   gs8_m16: if sdwidth=8 and mdwidth=16 generate
     sdo<= fifo_dout(0 to 7) when sa(awidth-1)='0' else fifo_dout(8 to 15);
     lsb_adr<= sa(awidth-1);
     mbe(1)<= ma(awidth-1) or seqAddrQ;  
      mbe(0)<= not ma(awidth-1);
   end generate;
        
   gs16m32: if sdwidth=16 and mdwidth=32 generate
      sdo<= fifo_dout(0 to 15) when sa(awidth-2)='0' else fifo_dout(16 to 31);
      lsb_adr<= sa(awidth-2); 
      mbe(3)<= '1' when (ma(awidth-2)='1' and sbe_q(1)='1') or seqAddrQ='1' else '0';
      mbe(2)<= '1' when (ma(awidth-2)='1' and sbe_q(0)='1') or (seqAddrQ='1' and ma(awidth-2)='0') else '0';  
      mbe(1)<= not ma(awidth-2) and sbe_q(1);
      mbe(0)<= not ma(awidth-2) and sbe_q(0);
   end generate;

   gs8m32: if sdwidth=8 and mdwidth=32 generate
      sdo<= fifo_dout(24 to 31) when sa(awidth-2 to awidth-1)="11" else 
            fifo_dout(16 to 23) when sa(awidth-2 to awidth-1)="10" else
            fifo_dout(8 to 15) when sa(awidth-2 to awidth-1)="01" else
            fifo_dout(0 to 7);     
      mbe(3)<= '1' when ma(awidth-2 to awidth-1)="11" or seqAddrQ='1' else '0';
      mbe(2)<= '1' when ma(awidth-2 to awidth-1)="10" or 
                   (seqAddrQ='1' and ma(awidth-2 to awidth-1)/="11") else '0';
      mbe(1)<= '1' when ma(awidth-2 to awidth-1)="01" or 
                   (seqAddrQ='1' and ma(awidth-2)='0') else '0';
      mbe(0)<= not ma(awidth-2) and not ma(awidth-1);
      lsb_adr<= sa(awidth-2) and sa(awidth-1);    
   end generate;
end arch;  


------------------------------------------------------------------------------------------------------
-- direct connection for write mdwidth>sdwidth
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity dwidth_conv_m_wr_direct is
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    buf_valid: out std_logic; -- valid data in buffer
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end dwidth_conv_m_wr_direct;

architecture arch of dwidth_conv_m_wr_direct is
  signal buf: std_logic_vector(0 to mdwidth-1);   -- internal write buffer
  signal buf_be: std_logic_vector( 0 to mdwidth/8 -1);
  signal buf_be_or: std_logic; -- or of all buf_be
  signal buf_be_LSB: std_logic; -- the LSB byte enable (for sdwidth=16 - 2 bytes are considered)
begin
   --   direct connection
   gwr_seqAddr0_size0: if use_seqAddr<=0 generate
         msel<= ssel;
         sack<= mack;
         buf_valid<= '0'; -- the is no buffer
         gs8m16: if sdwidth=8 and mdwidth=16 generate
           mdo(0 to 7)<= sdi; mdo(8 to 15)<= sdi;
           mbe(0)<= not sa(awidth-1); mbe(1)<= sa(awidth-1);
         end generate;
         gs16m32: if sdwidth=16 and mdwidth=32 generate
           mdo(0 to 15)<= sdi; mdo(16 to 31)<= sdi;
           mbe(0 to 1)<= sbe when sa(awidth-2)='0' else "00"; 
           mbe(2 to 3)<= sbe when sa(awidth-2)='1' else "00"; 
         end generate;
         gs8m32: if sdwidth=8 and mdwidth=32 generate
           mdo(0 to 7)<= sdi; mdo(8 to 15)<= sdi;
           mdo(16 to 23)<= sdi; mdo(24 to 31)<= sdi;
           mbe(0)<= not sa(awidth-2) and not sa(awidth-1); 
           mbe(1)<= not sa(awidth-2) and sa(awidth-1); 
           mbe(2)<= sa(awidth-2) and not sa(awidth-1);
           mbe(3)<= sa(awidth-2) and sa(awidth-1); 
         end generate;
   end generate; -- write logic - no fifo, use_seqAddr>0


   -- no write fifo but use write flip-flops
   -- use internal write buffer to concatenete several slave writes into a signle write - this reduces the master writes cycles
   gwr_seqAddr1_size0: if use_seqAddr>0 generate -- direct connection
         mbe<= buf_be;
         sack<= ssel and (not buf_be_or or -- write buffer is empty
           (seqAddrQ and not buf_be_LSB) or -- the LSB (the last transfer) is empty so can accept a new data when seqAddressing
           mack); -- remove this line for a slow transfers - this will shorten the ack path.
         buf_valid<= buf_be_or;    
         msel<= buf_be_or and (not seqAddrQ -- no seqAddressing or the last seqAddr transfer
                   or buf_be_LSB); -- the LSB is transfered so full word has been written

         
         gs8m16: if sdwidth=8 and mdwidth=16 generate
            buf_be_or<= buf_be(0) or buf_be(1);
            buf_be_LSB<= buf_be(1);
            
            process(arst, clk) begin
              if arst='1' then  
                  mdo<= (others=> '0');  
                  buf_be<= (others=> '0');
              elsif clk'event and clk='1' then
                if sack='1' and sa(awidth-1)='0' then -- MSB transfer
                    mdo(0 to 7)<= sdi;
                    buf_be(0)<='1';
                elsif mack='1' then 
                    buf_be(0)<='0';
                end if;
                if  sack='1' and sa(awidth-1)='1' then -- LSB transfer
                    mdo(8 to 15)<= sdi;
                    buf_be(1)<='1';
                elsif mack='1' then
                    buf_be(1)<='0';
                end if;
              end if;
            end process;
         end generate;

         gs16m32: if sdwidth=16 and mdwidth=32 generate
            buf_be_or<= buf_be(0) or buf_be(1) or buf_be(2) or buf_be(3);
            buf_be_LSB<= buf_be(2) or  buf_be(3);
            
            process(arst, clk) begin
              if arst='1' then mdo<= (others=> '0'); buf_be<= (others=> '0');
              elsif clk'event and clk='1' then
                if sa(awidth-2)='0' and sack='1' then -- the MSB 2 bytes transfer
                   if sbe(0)='1' then 
                      mdo(0 to 7)<= sdi(0 to 7);
                      buf_be(0)<='1';
                   end if;
                   if sbe(1)='1' then 
                      mdo(8 to 15)<= sdi(8 to 15);
                      buf_be(1)<='1';
                   end if;
                elsif mack='1' then
                    buf_be(0 to 1)<="00";
                end if;
                if sa(awidth-2)='1' and sack='1' then -- the LSB (2 bytes) slave transfer
                    if sbe(0)='1' then
                      mdo(16 to 23)<= sdi(0 to 7);
                      buf_be(2)<='1';
                    end if;
                    if sbe(1)='1' then
                      mdo(24 to 31)<= sdi(8 to 15);
                      buf_be(3)<='1';
                    end if;
                elsif mack='1' then
                    buf_be(2 to 3)<="00";
                end if;
              end if;
            end process;
       end generate;
         
             
      gs8m32: if sdwidth=8 and mdwidth=32 generate
         buf_be_or<= buf_be(0) or buf_be(1) or buf_be(2) or buf_be(3);
         buf_be_LSB<=  buf_be(3);
          
            process(arst, clk) begin
              if arst='1' then mdo<= (others=> '0'); buf_be<= (others=> '0');
              elsif clk'event and clk='1' then
                if sa(awidth-2 to awidth-1)="00" and sack='1' then -- the MSB trnasfer
                      mdo(0 to 7)<= sdi;
                      buf_be(0)<='1';
                elsif mack='1' then
                    buf_be(0)<= '0';
                end if;
                if sa(awidth-2 to awidth-1)="01" and sack='1' then 
                      mdo(8 to 15)<= sdi;
                      buf_be(1)<='1';
                elsif mack='1' then
                    buf_be(1)<= '0';
                end if;
                if sa(awidth-2 to awidth-1)="10" and sack='1' then 
                      mdo(16 to 23)<= sdi;
                      buf_be(2)<='1';
                elsif mack='1' then
                    buf_be(2)<= '0';
                end if;
                if sa(awidth-2 to awidth-1)="11" and sack='1' then 
                      mdo(24 to 31)<= sdi;
                      buf_be(3)<='1';
                elsif mack='1' then
                    buf_be(3)<= '0';
                end if;
              end if;
            end process;
       end generate;
          
   end generate; -- if use_seqAddr>0
end arch;    

------------------------------------------------------------------------------------------------------
-- write mdwidth>sdwidth with fifo
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity dwidth_conv_m_wr_fifo is
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    fifo_size: integer:= 4; -- fifo address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    buf_valid: out std_logic; -- valid data in buffer
    buf_almost_empty: out std_logic; -- write byffer can be empty if in this clk cycle data are transfered out
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end dwidth_conv_m_wr_fifo;

architecture arch of dwidth_conv_m_wr_fifo is
 component dwidth_conv_m_wr_direct
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    buf_valid: out std_logic; -- valid data in buffer
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end component;
component fifo
  generic( adr_width: integer:= 4; -- internal memory address width - 4 is preferable for distributed rams, 0- direct connection
	  data_width: integer:= 8); -- data bus width
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
	rst: in std_logic; -- synchronous reset for the whole device
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    half_full: out std_logic; -- fifo is half full
	-- data input signals (slave)
	datI_I: in std_logic_vector(0 to data_width-1);
	stbI_I: in std_logic; -- strobe similar as OPB_select (master ready to transfer data)
	ackI_O: out std_logic; -- acknowledge
	-- data output signals (master)
	datO_O: out std_logic_vector(0 to data_width-1);
	stbO_O: out std_logic;
	ackO_I: in std_logic);
  end component;


  -- data width converter signals
  signal dconv_buf_valid, dconv_msel, dconv_mack: std_logic;
  signal dconv_mdo: std_logic_vector(0 to mdwidth-1);
  signal dconv_mbe: std_logic_vector(0 to mdwidth/8 -1);
  -- fifo signals
  signal fifo_din, fifo_dout: std_logic_vector(0 to mdwidth + mdwidth/8 -1); -- data and byte enable logic
  signal fifo_almost_empty, fifo_stbO_O: std_logic;
begin
 dwc: dwidth_conv_m_wr_direct
  generic map(sdwidth=> sdwidth, mdwidth=> mdwidth, awidth=> awidth, use_seqAddr=> use_seqAddr)
  port map(clk=> clk, arst=> arst,
    -- slave ifterface (input)
    sdi=> sdi, sbe=> sbe, sa=> sa, sseqAddr=> sseqAddr, seqAddrQ=> seqAddrQ, 
	ssel=> ssel, sack=> sack,
	-- master interface (output)
    buf_valid=> dconv_buf_valid, mdo=> dconv_mdo, mbe=> dconv_mbe, 
	msel=> dconv_msel, mack=> dconv_mack);

 fi: fifo
  generic map(adr_width=> fifo_size, data_width=> mdwidth + mdwidth/8)
  port map (clk=> clk, arst=> arst, rst=> '0',
    almost_full=> open, almost_empty=> fifo_almost_empty, full=> open, half_full=> open,
	-- data input signals (slave)
	datI_I=> fifo_din, stbI_I=> dconv_msel, ackI_O=> dconv_mack,
	-- data output signals (master)
	datO_O=> fifo_dout, stbO_O=> fifo_stbO_O, ackO_I=> mack);

  fifo_din<= dconv_mdo & dconv_mbe; -- data and byte enable are stored in the fifo
  mdo<= fifo_dout(0 to mdwidth-1);
  mbe<= fifo_dout(mdwidth to mdwidth + mdwidth/8 -1);
  msel<= fifo_stbO_O;
  buf_valid<= dconv_buf_valid or fifo_stbO_O;
  buf_almost_empty<= not dconv_buf_valid and fifo_almost_empty;    
end arch;



------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity dwidth_conv_m is
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1; -- 0- ignore seqAddr logic, 1- implement seqAddr
    fifo_wr_size: integer:= 4; 
      -- <0 - do not use write logic at all
      -- 0 - widthout fifo (direct connection
      -- >0 - with fifo- fifo size 2^fifo_wr_size
    fifo_rd_size: integer:= 4); -- read fifo size - see fifo_wr_size 
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    srnw: in std_logic; -- read not write
    sseqAddr: in std_logic; -- sequential addressing
	ssel: in std_logic; -- OPB select signal
	sack: out std_logic; -- slave acknowledge
	-- master interface (output)
    mdi: in std_logic_vector(0 to mdwidth-1); -- data input
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
    ma: out std_logic_vector(0 to awidth-1); -- address bus
    mrnw: out std_logic; -- read not write
	msel: out std_logic; -- OPB select signal
    mseqAddr: out std_logic;
	mack: in std_logic);
end dwidth_conv_m;


architecture dwidth_conva of dwidth_conv_m is
 component dwidth_conv_m_wr_direct
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    buf_valid: out std_logic; -- valid data in buffer
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
   end component;

component dwidth_conv_m_rd_direct
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    mdi: in std_logic_vector(0 to mdwidth-1); -- data input
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end component;

component dwidth_conv_m_rd_fifo
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    fifo_size: integer:= 4; -- address width of the fifo
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    ma: in std_logic_vector(0 to awidth-1); -- address bus
    ma_valid: in std_logic; -- valid master address
    mdi: in std_logic_vector(0 to mdwidth-1); -- data input
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
end component;
component dwidth_conv_m_wr_fifo
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    fifo_size: integer:= 4; -- fifo address width
    use_seqAddr: integer:= 1); -- 0- ignore seqAddr logic, 1- implement seqAddr
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    sseqAddr: in std_logic; -- sequential addressing
    seqAddrQ: in std_logic; -- sequential addressing Q (detects low value on seqAddr
	ssel: in std_logic; -- OPB select signal
	sack: buffer std_logic; -- slave acknowledge
	-- master interface (output)
    buf_valid: out std_logic; -- valid data in buffer
    buf_almost_empty: out std_logic; -- write byffer can be empty if in this clk cycle data are transfered out
    mdo: out std_logic_vector(0 to mdwidth-1); -- data output
    mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
	msel: out std_logic; -- OPB select signal
	mack: in std_logic);
  end component;

  signal ma_valid, ma_valid_rd: std_logic; -- valid ma bus
  signal ma_q: std_logic_vector(0 to awidth-1); -- master address (counter output)
  signal wr_ma_q: std_logic; -- load address counter
  signal mseqAddr1, seqAddrQ: std_logic;
  signal buf_wr_valid, buf_wr_almost_empty, ssel_rd, ssel_wr: std_logic; -- valid data in the write buffer
  signal mrnw1, srnw_q, msel_wr, msel_rd, sack_rd, sack_wr: std_logic; -- output signal seperate for reads and writes
  signal mack_wr, mack_rd: std_logic; 
  signal mbe_rd, mbe_wr: std_logic_vector(0 to mdwidth/8 -1);
begin
  mseqAddr<= mseqAddr1; -- or mseqAddr_wr;  

  msel<= msel_wr or msel_rd;
  sack<= sack_rd or sack_wr;
  mbe<= mbe_rd when mrnw1='1' else mbe_wr;
  mrnw<= mrnw1;
  
  -- no write logic
  g_wr0: if fifo_wr_size<0 generate -- do not use write logic
         mdo<= (others=> '0');
         mbe_wr<= (others=> '0');
         msel_wr<= '0';
         sack_wr<= '0';
		 buf_wr_valid<= '0'; -- write buffer is always empty (no write)
  end generate;           
  -- no read logic
  g0: if fifo_rd_size<0 generate
      sdo<= (others=>'0');
      mbe_rd<= (others=> '0');
      msel_rd<= '0';
      sack_rd<= '0';
   end generate;

   
--------------------------------------------------------------------------           
-- WRITE LOGIC  -- direct connection
  gwr_size0: if fifo_wr_size=0 generate
    dwr_direct: dwidth_conv_m_wr_direct
     generic map(sdwidth=> sdwidth, mdwidth=> mdwidth, awidth=> awidth, use_seqAddr=>use_seqAddr)
     port map (clk=> clk, arst=> arst,
      -- slave ifterface (input)
      sdi=> sdi, sbe=>sbe, sa=> sa, sseqAddr=> sseqAddr, 
      seqAddrQ=> seqAddrQ, ssel=> ssel_wr, sack=> sack_wr,
	  -- master interface (output)
      buf_valid=> buf_wr_valid, mdo=> mdo, mbe=> mbe_wr,
      msel=> msel_wr, mack=> mack_wr);

      mack_wr<= mack and not mrnw1;
      ssel_wr<= ssel and not srnw;
      buf_wr_almost_empty<= '1';
   end generate;   
  
   gwr_size1: if fifo_wr_size > 0 generate
     fwr: dwidth_conv_m_wr_fifo 
       generic map(sdwidth=> sdwidth, mdwidth=> mdwidth,
        awidth=> awidth, fifo_size=> fifo_wr_size, use_seqAddr=> use_seqAddr)
       port map(clk=> clk, arst=> arst,
        -- slave ifterface (input)
       sdi=> sdi, sbe=> sbe, sa=> sa,
       sseqAddr=> sseqAddr, seqAddrQ=> seqAddrQ, ssel=> ssel_wr, sack=> sack_wr,
	   -- master interface (output)
       buf_valid=> buf_wr_valid, buf_almost_empty=> buf_wr_almost_empty,
       mdo=> mdo, mbe=> mbe_wr, msel=> msel_wr, mack=> mack_wr);

       mack_wr<= mack and not mrnw1;
       ssel_wr<= ssel and not srnw;
   end generate;   

---------------------------------------------------------------------------------------------   
 -- READ LOGIC direct connection
  grd_size0: if fifo_rd_size = 0 generate
    drd_direct: dwidth_conv_m_rd_direct
     generic map(sdwidth=> sdwidth, mdwidth=> mdwidth, awidth=> awidth, use_seqAddr=>use_seqAddr)
     port map (clk=> clk, arst=> arst,
        -- slave ifterface (input)
        sdo=> sdo, sbe=> sbe, sa=> sa, sseqAddr=> sseqAddr, seqAddrQ=> seqAddrQ, 
	    ssel=> ssel_rd, sack=> sack_rd,
	    -- master interface (output)
       mdi=> mdi, mbe=> mbe_rd, msel=> msel_rd, mack=> mack_rd);

       mack_rd<= mack and mrnw1;
       ssel_rd<= ssel and srnw;
   end generate;
   
   -- read logic - use fifo
  grd_size1: if fifo_rd_size>0 generate
    drd_fifo: dwidth_conv_m_rd_fifo
    generic map (sdwidth=> sdwidth, mdwidth=> mdwidth, awidth=> awidth, 
        fifo_size=> fifo_rd_size, use_seqAddr=> use_seqAddr)
    port map (clk=> clk, arst=> arst,
       -- slave ifterface (input)
       sdo=> sdo, sbe=> sbe, sa=> sa, sseqAddr=> sseqAddr, seqAddrQ=> seqAddrQ,
	   ssel=> ssel_rd, sack=> sack_rd,
	   -- master interface (output)
       ma=> ma_q, ma_valid=> ma_valid_rd, mdi=> mdi, mbe=> mbe_rd,
	  msel=> msel_rd, mack=> mack_rd);
      
      ma_valid_rd<= ma_valid and srnw_q;
      mack_rd<= mack and mrnw1;
      ssel_rd<= ssel and srnw;
   end generate;
   
-----------------------------------------------------------------------------------------        
-----------------------------------------------------------------------------------        
-- Adress counter logic  
  g_seqAddr1: if use_seqAddr>0 generate
    process(clk,arst) begin
      if arst='1' then seqAddrQ<= '0';
      elsif clk'event and clk='1' then
        if seqAddrQ='1' then 
          seqAddrQ<= sseqAddr; -- detect every low value on seqAddr
        elsif mseqAddr1='0' and ssel='1' then -- the low state on sseqAddr must be transmitted to the mseqAddr therefore mseqAddr1='0'
          seqAddrQ<= sseqAddr;
        end if;
      end if;
     end process;
  end generate;
  
  
  gseqAddr0: if use_seqAddr<=0 generate
      seqAddrQ<= '0';
  end generate;
  
  mseqAddr1<= '0' when use_seqAddr<=0 else
    sseqAddr when mrnw1='1' and fifo_rd_size<=0 else
    seqAddrQ when mrnw1='1' and fifo_rd_size>0 else
    not buf_wr_almost_empty or seqAddrQ; -- for write
  
  -- generate internal address counter
  g_maq: if fifo_rd_size>0 or fifo_wr_size>0 or (fifo_wr_size=0 and use_seqAddr>0) generate
   process(arst, clk) begin
    if arst='1' then ma_q<= (others=>'0'); ma_valid<= '0'; srnw_q<= '0';
    elsif clk'event and clk='1' then
      if wr_ma_q= '1' then
        ma_q<= sa; -- latch the address from the slave
        ma_valid<= '1';
        srnw_q<= srnw; -- remember the cycle type
      else
        if (srnw_q='0' and ((buf_wr_almost_empty='1' and mack='1' and seqAddrQ='0') or -- write buffer will be empty in the next clk cycle
            (buf_wr_valid='0' and seqAddrQ= '0'))) or -- write buffer is empty and not seq Addressing
            ( srnw_q='1' and (seqAddrQ='0' or sseqAddr='0')) then -- read logic - stop reading 
              ma_valid<= '0';
        end if;          
        if mack='1' and use_seqAddr>0 then -- imcrease address by 1
           if mdwidth=16 then 
             ma_q(awidth-1)<= '0'; ma_q(0 to awidth-2)<= ma_q(0 to awidth-2)+1;
           elsif mdwidth= 32 then 
              ma_q(awidth-2 to awidth-1)<= "00"; ma_q(0 to awidth-3)<= ma_q(0 to awidth-3) + 1;
           end if;
        end if;
      end if;
    end if;
   end process;
   
   -- load ma_q counter
   wr_ma_q<= '1' when ssel='1' and seqAddrQ='0' 
    -- and sack_rd='0'  and mack_rd='0' -- read buffer is empty (seqAddr=0 so do not read any more)
    and (buf_wr_valid='0' or (buf_wr_almost_empty='1' and mack='1')) -- write fifo empty or will be empty after this clk
    else '0';
  end generate; -- use internal address counter
  
  
  mrnw1<= srnw_q when (fifo_wr_size/=0 or (fifo_wr_size=0 and use_seqAddr>0)) and fifo_rd_size/=0 else -- both read and write uses address counter
  not buf_wr_valid and srnw when fifo_rd_size=0 else -- direct read (no address counter for reading) and write using address counter
  srnw;
          
      
  -- address multiplexer
  ma<= sa when (mrnw1='1' and fifo_rd_size= 0) or -- read 
            (mrnw1='0' and fifo_wr_size=0 and use_seqAddr<=0) --write 
       else ma_q; -- use address from address counter  
end dwidth_conva;
