-- opb2opb_dwidth module
-- this module can transfer data from one OPB Bus to another OPB bus 
-- with data width conversion
-- allowed data widths: 8, 16, 32
-- by Erenst Jamro 
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset signal

entity opb2opb_dwidth is
  generic (c_opb_dwidth: integer:= 32; -- OPB data bus width (same for all buses - use c_dwidthM and c_dwidthS for real data widths
           c_opb_awidth: integer:= 32; -- OPB address bus width
		       c_sdwidth: integer:= 8; -- real OPB data width of the Slave side
           c_mdwidth: integer:= 16; -- real OPB data width of the Master size
           c_fifo_wr_awidth: integer:= 4; -- fifo address width:
             -- <0 - do not implement write logic at all
             -- 0 - direct connection (no fifo)
             -- 4 fifo address width (2**c_fifo_wr_awidth location) - should be 4 (optimal for small distributed BRAMs)
		       c_fifo_rd_awidth: integer:= 0; -- fifo for reading - see c_fifo_wr_awidth for more information
           c_use_seqAddr: integer:= 0; -- use sequential addressing
             -- 0 - do not use seq Addr
             -- 1 - use seqAddr
           c_use_toutSup: integer:= 0; -- =0 - do not generate OPB_toutSup, =1- generate OPB_toutSup when this device is selected
		       c_speedrate : integer:= 0; -- should be 0 unless the minimum period time condition is not met
           c_transfer_size: integer:= 4; -- max transfer size when bus is locked (valid only when seqAddr>0)
           C_MSBADDR   : std_logic_vector(0 to 31) := X"FFFF_FFFF"; -- MSBs of master address - used only when fifferent form FFFF_FFFF
           C_BASEADDR   : std_logic_vector(0 to 31) := X"0000_0000"; -- only single address region is decoded
           C_HIGHADDR   : std_logic_vector(0 to 31) := X"0000_01FF");
	port (OPB_clk: in STD_LOGIC; 
	    OPB_rst: in STD_LOGIC;
		-- OPB Slave interface
       SOPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
       SOPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
       SOPB_RNW     : in std_logic;
       SOPB_select  : in std_logic;
       SOPB_seqAddr : in std_logic;
       SOPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_errAck  : out std_logic;
       Sl_retry   : out std_logic;
       Sl_toutSup : out std_logic;
       Sl_xferAck : out std_logic;
       -- OPB Master interface
	   M_ABus : out STD_LOGIC_VECTOR(0 TO c_opb_awidth-1); -- is already gated by M_Select
	   M_DBus : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
       M_BE : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth/8 -1); 
	   M_RNW : out STD_LOGIC;
	   M_select : out STD_LOGIC;
	   M_seqAddr : out STD_LOGIC;
	   M_busLock : out STD_LOGIC;
	   M_request : out STD_LOGIC;
	   MOPB_MGrant: in STD_LOGIC;
	   MOPB_pendReq: in STD_LOGIC;
	   MOPB_DBus : in STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
	   MOPB_errAck : in STD_LOGIC;
	   MOPB_retry : in STD_LOGIC;
	   MOPB_timeout: in STD_LOGIC;
	   MOPB_xferAck : in STD_LOGIC);
end opb2opb_dwidth;

architecture opb2opb_dwidth_arch of opb2opb_dwidth is
-- functions and components declarations
function FindAWidth(base_adr: std_logic_vector; high_adr: std_logic_vector) 
    return integer is
  begin
	for i in C_OPB_AWIDTH-1 downto 0 loop
       if base_adr(i) = high_adr(i) then
		  return C_OPB_AWIDTH-i-1;
	  end if;
	end loop;
	return C_OPB_AWIDTH;
end FindAWidth;                 

component ROC -- reset on Configuration
   port (O : out std_logic);
  end component;

component dwidth_conv_m
  generic(sdwidth: integer:= 16; -- width of the input bus (Slave)
    mdwidth: integer:= 8; -- width of the output bus (master)
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1; -- 0- ignore seqAddr logic, 1- implement seqAddr
    fifo_wr_size: integer:= 4; 
    fifo_rd_size: integer:= 4);
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to sdwidth-1); -- data input
    sdo: out std_logic_vector(0 to sdwidth-1); -- data output
    sbe: in std_logic_vector(0 to sdwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    srnw: in std_logic; -- read not write
    sseqAddr: in std_logic; -- sequential addressing
	ssel: in std_logic; -- OPB select signal
	sack: out std_logic; -- slave acknowledge
	-- master interface (output)
  mdi: in std_logic_vector(0 to mdwidth-1); -- data input
  mdo: out std_logic_vector(0 to mdwidth-1); -- data output
  mbe: out std_logic_vector(0 to mdwidth/8 -1); -- byte enable
  ma: out std_logic_vector(0 to awidth-1); -- address bus
  mrnw: out std_logic; -- read not write
	msel: out std_logic; -- OPB select signal
  mseqAddr: out std_logic;
	mack: in std_logic);
end component;
component fifo_adr -- the same data width
  generic(dwidth: integer:= 16; -- data bus width
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1; -- 0- ignore seqAddr logic, 1- implement seqAddr
    fifo_wr_size: integer:= 4; 
      -- <0 - do not use write logic at all
      -- 0 - widthout fifo (direct connection)
      -- >0 - with fifo- fifo size 2^fifo_wr_size
    fifo_rd_size: integer:= 4); -- read fifo size - see fifo_wr_size 
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to dwidth-1); -- data input
    sdo: out std_logic_vector(0 to dwidth-1); -- data output
    sbe: in std_logic_vector(0 to dwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    srnw: in std_logic; -- read not write
    sseqAddr: in std_logic; -- sequential addressing
	ssel: in std_logic; -- OPB select signal
	sack: out std_logic; -- slave acknowledge
	-- master interface (output)
  mdi: in std_logic_vector(0 to dwidth-1); -- data input
  mdo: out std_logic_vector(0 to dwidth-1); -- data output
  mbe: out std_logic_vector(0 to dwidth/8 -1); -- byte enable
  ma: out std_logic_vector(0 to awidth-1); -- address bus
  mrnw: out std_logic; -- read not write
	msel: out std_logic; -- OPB select signal
    mseqAddr: out std_logic;
	mack: in std_logic);
  end component;

component opb_master is
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width the MSB of OPB_DBus='0'
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal address bus width
    transfer_size: integer:= 4); -- max transfer size when bus is locked (valid only when seqAddr>0)
  port(clk, arst: std_logic; --clock and global asynchronous reset
		-- OPB Master interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO opb_dwidth/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    M_request : out STD_LOGIC;
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
        -- internal simplified bus
      di: in std_logic_vector(0 to dwidth-1); -- data input
      do: out std_logic_vector(0 to dwidth-1); -- data output
      be: in std_logic_vector(0 to dwidth/8 -1); -- byte enable
      a: in std_logic_vector(0 to awidth-1); -- address bus
      msb_adr: in std_logic_vector(0 to opb_awidth-1):=(others=>'0'); -- the MSB of the OPB address are specified by the msb_adr
      rnw: in std_logic; -- read not write
	    sel: in std_logic; -- OPB select signal
      seqAddr: in std_logic;
	    ack: out std_logic);

end component;
   
-----------------------------------------------------------------------------
  constant awidth: integer:=  FindAWidth(c_baseaddr, c_highaddr);  -- address width when address spac is considered
   
  signal arst: std_logic; -- Global set/reset
  signal my_adr, my_adrQ: std_logic; -- the slave device address is selected
  signal dev_select: std_logic; -- slave device is selected
  signal xferAck, xferAckQ: std_logic; -- internal SL_xferAck
  -- data width converter to OPB master signals- signals name same as for dwidth_conv
  signal mdi, mdo: std_logic_vector(0 to c_mdwidth-1); -- data bus
  signal mbe: std_logic_vector(0 to c_mdwidth/8-1); -- byte enable
  signal ma: std_logic_vector(0 to awidth-1);
  signal msb_adr: std_logic_vector(0 to c_opb_awidth-1);
  signal mrnw, mack, msel, mseqAddr: std_logic;
  signal sdo: std_logic_vector(0 to c_sdwidth-1); -- slave data out
  
begin
  -- decode the address of this device
  ga1: if C_OPB_AWIDTH > awidth generate
	my_adr<= '1' when SOPB_ABus(0 to C_OPB_AWIDTH - awidth -  1) 
        = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
       else '0';
  end generate;
  
  -- full address space is used	(adr_space_width=0 when highaddr - baseaddr +1 overflows)
  ga0: if C_OPB_AWIDTH = awidth generate
     my_adr<= '1';
  end generate;
  
  -- SLAVE SIDE LOGIC
	-- direct logic
  gspeed0: if c_speedrate = 0 generate
    dev_select<= SOPB_select and my_adr;
	Sl_DBus(0 to c_sdwidth-1)<= sdo when dev_select='1' and SOPB_RNW='1' else (others=> '0');
	Sl_xferAck<= xferAck;
  end generate;	
  
  -- address decoder is pipelined (one clock delay for not seqAddres) only one additional flip-flop
  gspeed1: if c_speedrate = 1 generate 
	process(arst, OPB_clk) begin
	  if arst = '1' then my_adrq<= '0';
	  elsif OPB_clk'event and OPB_clk='1' then
		if my_adr='1' and ( xferAck='0' or (SOPB_SeqAddr='1' and c_use_seqAddr>0)) then
           my_adrq<='1';
		else my_adrq<='0';
	    end if;
	  end if;
	end process;
    dev_select<= SOPB_select and my_adrq;
	Sl_DBus(0 to c_sdwidth-1)<= sdo when dev_select='1' and SOPB_RNW='1' else (others=> '0');
	Sl_xferAck<= xferAck;
  end generate;
  
  -- xferAck is pipelined (one clock delay (even with seqAddr) and data-width additional flip-flops
  gspeed2: if c_speedrate=2 generate
    dev_select<= SOPB_select and my_adr and not xferAckQ;
	Sl_xferAck<= xferAckQ;
	
	process(OPB_clk, arst) begin
	  if arst= '1' then Sl_DBus(0 to c_sdwidth-1)<= (others=> '0'); xferAckQ<= '0';
	  elsif OPB_clk'event and OPB_clk='1' then
	    if my_adr='1' and SOPB_select='1' and  SOPB_RNW='1' then
	      Sl_DBus(0 to c_sdwidth-1)<= sdo;
		else Sl_DBus(0 to c_sdwidth-1)<= (others=> '0');
	    end if;
		if my_adr='1' and SOPB_select='1' and xferAck='1' and xferAckQ='0' then
		  xferAckQ<= '1';
		else xferAckQ<= '0';
		end if;
	  end if;
	end process;
  end generate;
    

   gsd: if c_sdwidth< c_opb_dwidth generate
      Sl_DBus(c_sdwidth to c_opb_dwidth-1)<= (others=> '0');
   end generate;
        
---------------------------------------------------------------------------
-- data width conversion
  gdm: if c_mdwidth>c_sdwidth generate
    m: dwidth_conv_m
     generic map(sdwidth=> c_sdwidth, mdwidth=> c_mdwidth, 
      awidth=> awidth, use_seqAddr=> c_use_seqAddr, 
      fifo_rd_size=> c_fifo_rd_awidth, fifo_wr_size=> c_fifo_wr_awidth)
    port map (clk=> OPB_clk, arst=> arst, 
    -- slave ifterface (input)
    sdi=> SOPB_DBus(0 to c_sdwidth-1), sdo=> sdo,
    sbe=> SOPB_BE(0 to c_sdwidth/8 -1),
    sa=> SOPB_ABus(c_opb_awidth-awidth to c_opb_awidth-1), 
    srnw=> SOPB_RNW, sseqAddr=> SOPB_seqAddr,
	  ssel=> dev_select, sack=> xferAck,
	-- master interface (output)
    mdi=> mdi, mdo=> mdo, mbe=> mbe, ma=> ma,
    mrnw=> mrnw, msel=> msel, mseqAddr=> mseqAddr, mack=> mack);
   
  end generate; -- mdwidth > sdwidth

  eq: if c_mdwidth= c_sdwidth generate
     fa:  fifo_adr 
    generic map(dwidth=> c_sdwidth, awidth=> awidth, use_seqAddr=> c_use_seqAddr,
      fifo_wr_size=> c_fifo_wr_awidth, fifo_rd_size=> c_fifo_rd_awidth)
    port map (clk=> OPB_clk, arst=> arst,
      -- slave ifterface (input)
      sdi=> SOPB_DBus(0 to c_sdwidth-1), sdo=> sdo,
      sbe=> SOPB_BE(0 to c_sdwidth/8 -1), 
      sa=> SOPB_ABus(c_opb_awidth-awidth to c_opb_awidth-1),
      srnw=> SOPB_RNW, sseqAddr=> SOPB_seqAddr,
	    ssel=> dev_select, sack=> xferAck,
	    -- master interface (output)
      mdi=> mdi, mdo=> mdo, mbe=> mbe, ma=> ma,
      mrnw=> mrnw, msel=> msel, mseqAddr=> mseqAddr, mack=> mack);
  end generate;
   -------------------------------------------------------------------------------------------
   -- master logic
   mstr: opb_master
     generic map(opb_dwidth=> c_opb_dwidth, dwidth=> c_mdwidth, opb_awidth=> c_opb_awidth,
       awidth=> awidth, transfer_size=> c_transfer_size)
     port map(clk=> opb_clk, arst=> arst,
		-- OPB Master interface
	    M_ABus=> M_ABus, M_BE=> M_BE, M_RNW=> M_RNW,
	    M_select=> M_select, M_seqAddr=> M_seqAddr, M_busLock=> M_busLock,
	    M_DBus=> M_DBus, M_request=> M_request, 
      OPB_MGrant=> MOPB_MGrant, OPB_pendReq=> MOPB_pendReq,
	    OPB_DBus=> MOPB_DBus, OPB_errAck=> MOPB_errAck, OPB_retry=> MOPB_retry,
	    OPB_timeout=> MOPB_timeout, OPB_xferAck=> MOPB_xferAck,
        -- internal simplified bus
        di=> mdo, do=> mdi, be=> mbe, a=> ma, msb_adr=> msb_adr,
        rnw=> mrnw, sel=> msel, seqAddr=> mseqAddr, ack=> mack);
   
    
   Sl_toutSup<= dev_select when c_use_toutSup>0 else '0';
   Sl_retry<= '0';
   Sl_errAck<= '0';
       
                                                        
  -- select master msb_adr the same as slave base address
  ma0: if c_msbaddr = x"FFFFFFFF" generate -- = FFFF_FFFF
    msb_adr<= c_baseaddr;
  end generate;
  -- the master msb_Adr is defined by c_msb_addr
  ma1: if c_msbaddr /= x"FFFFFFFF" generate
    msb_adr<= c_msbaddr;
  end generate;
    
 -- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
 global_reset: ROC
	port map (O=> arst);

end opb2opb_dwidth_arch;