-- master device that allows a simplified OPB bus to be connected to OPB bus as a master
-- author: Ernest Jamro jamro@agh.edu.pl

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity opb_master is
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width the MSB of OPB_DBus='0'
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal address bus width
    transfer_size: integer:= 4); -- max transfer size when bus is locked (valid only when seqAddr>0)
  port(clk, arst: std_logic; --clock and global asynchronous reset
		-- OPB Master interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO opb_dwidth/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    M_request : out STD_LOGIC;
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
        -- internal simplified bus
        di: in std_logic_vector(0 to dwidth-1); -- data input
        do: out std_logic_vector(0 to dwidth-1); -- data output
        be: in std_logic_vector(0 to dwidth/8 -1); -- byte enable
        a: in std_logic_vector(0 to awidth-1); -- address bus
        msb_adr: in std_logic_vector(0 to opb_awidth-1):=(others=>'0'); -- the MSB of the OPB address are specified by the msb_adr
        rnw: in std_logic; -- read not write
	    sel: in std_logic; -- OPB select signal
        seqAddr: in std_logic; -- seqAddressing - a (address) is incremented - block transfer - similar logic assert OPB_seqAddr
	    ack: out std_logic);
end opb_master;

architecture opb_master_arch of opb_master is
function log2(arg: integer) return integer is
  variable ret: integer:= -1;
  begin
    for i in 0 to 32 loop
      if 2**i = arg then ret:= i; exit; end if;
    end loop;
   return ret;
 end log2;
 function MAX(arg0: integer; arg1: integer) return integer is
  variable ret: integer;
  begin
    if arg0>=arg1 then 
      return arg0;
    else 
      return arg1;
    end if;
 end MAX;
  constant transfer_size1: integer:= MAX(1, transfer_size);
  signal bus_granted: std_logic; -- OPB bus is granted
  signal OPB_sel: std_logic; -- internal OPB select signal
  signal transfer_count: integer range 0 to transfer_size1-1; -- how many transfer since bus master
  signal pendReq_q: std_logic; -- another device also request the bus - flip-flop output
  signal seqAddr1: std_logic; -- similar like seqAddr but OPB logic is added
  signal transfer_last: std_logic; -- last transfer
begin
  -- bus_granted logic
  pbus_grant: process(clk, arst) begin
	if arst='1' then bus_granted<= '0';
	elsif clk'event and clk='1' then
       if OPB_MGrant='1' then
         bus_granted<= '1';
	   elsif OPB_retry='1' or OPB_timeout='1' or OPB_errAck='1' -- error condition
           or sel='0' -- device is not ready for data transfer 
           or (OPB_xferAck='1' and (seqAddr='0' -- do not use sequencial addressing or not seqAddr cycle
               or (transfer_last='1' and pendReq_q='1'))) then -- relinquesh bus only when another master requensts the bus
		 bus_granted<= '0';
	   end if;
	end if;
  end process;								   
  
  transfer_last<= '1' when transfer_count= transfer_size1-1 else '0';
  

  ptransfer_count: process(clk, arst) begin
    if arst='1' then transfer_count<= 0; pendReq_q<= '0';
    elsif clk'event and clk='1' then
	  pendReq_q<= OPB_pendReq or (pendReq_q and bus_granted); -- keep pendReq_q high until the bus is not granted - result of the OPB_bus arbiter error
      if bus_granted='0' then
         transfer_count<= 0;
      elsif transfer_last='0' and OPB_xferAck='1' then
         transfer_count<= transfer_count + 1; -- counter with saturation
      end if;
    end if;
  end process;		
  
  OPB_sel<= bus_granted and sel;
  M_select<= OPB_sel;
    
  gd: if dwidth<opb_dwidth generate
     M_BE(dwidth/8 to opb_dwidth/8-1)<= (others=>'0');
     M_DBus(dwidth to opb_dwidth-1)<= (others=>'0'); -- the LSBs are always zero
  end generate;

  M_BE(0 to dwidth/8-1)<= be when bus_granted='1' else (others=>'0');
  M_DBus(0 to dwidth-1)<= di when bus_granted='1' and rnw='0' else (others=>'0');
  M_RNW<= rnw when bus_granted='1' else '0';
  seqAddr1<= '0' when bus_granted='0' else
   	  '0' when transfer_last='1' and pendReq_q='1' else -- when the last cycle
	  seqAddr and sel;
  M_seqAddr<= seqAddr1;
  M_busLock<= seqAddr1;

  M_request<= '0' when sel='0' else
	     '1' when bus_granted='0' else
		   '1' when pendReq_q='0' else '0' ;
  ack<= OPB_xferAck and bus_granted and sel; -- sometimes ack is genereted even if the sel is not longer active
  do<= OPB_DBus(0 to dwidth-1);
  
  M_ABus(opb_awidth-awidth to opb_awidth-1)<= a when bus_granted='1' else (others=>'0');  
  ga: if opb_awidth > awidth generate
    M_ABus(0 to opb_awidth-awidth-1)<= msb_adr(0 to opb_awidth-awidth-1) 
      when bus_granted='1' else (others=>'0');
  end generate;
  
end opb_master_arch;