-- fifo for simplified OPB, with address logic
-- author: Ernest Jamro jamro@agh.edu.pl

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////

  

------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity fifo_adr is
  generic(dwidth: integer:= 16; -- data bus width
    awidth: integer:= 32; -- address width
    use_seqAddr: integer:= 1; -- 0- ignore seqAddr logic, 1- implement seqAddr
    fifo_wr_size: integer:= 4; 
      -- <0 - do not use write logic at all
      -- 0 - widthout fifo (direct connection)
      -- >0 - with fifo- fifo size 2^fifo_wr_size
    fifo_rd_size: integer:= 4); -- read fifo size - see fifo_wr_size 
  port (clk, arst: in std_logic; -- clock and global asynchronous set reset
    -- slave ifterface (input)
    sdi: in std_logic_vector(0 to dwidth-1); -- data input
    sdo: out std_logic_vector(0 to dwidth-1); -- data output
    sbe: in std_logic_vector(0 to dwidth/8 -1); -- byte enable
    sa: in std_logic_vector(0 to awidth-1); -- address bus
    srnw: in std_logic; -- read not write
    sseqAddr: in std_logic; -- sequential addressing
	ssel: in std_logic; -- OPB select signal
	sack: out std_logic; -- slave acknowledge
	-- master interface (output)
    mdi: in std_logic_vector(0 to dwidth-1); -- data input
    mdo: out std_logic_vector(0 to dwidth-1); -- data output
    mbe: out std_logic_vector(0 to dwidth/8 -1); -- byte enable
    ma: out std_logic_vector(0 to awidth-1); -- address bus
    mrnw: out std_logic; -- read not write
	msel: out std_logic; -- OPB select signal
    mseqAddr: out std_logic;
	mack: in std_logic);
end fifo_adr;

architecture arch of fifo_adr is
  component fifo
  generic( adr_width: integer:= 4; -- internal memory address width - 4 is preferable for distributed rams, 0- direct connection
	  data_width: integer:= 8); -- data bus width
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
	rst: in std_logic; -- synchronous reset for the whole device
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    half_full: out std_logic; -- fifo is half full
	-- data input signals (slave)
	datI_I: in std_logic_vector(0 to data_width-1);
	stbI_I: in std_logic; -- strobe similar as OPB_select (master ready to transfer data)
	ackI_O: out std_logic; -- acknowledge
	-- data output signals (master)
	datO_O: out std_logic_vector(0 to data_width-1);
	stbO_O: out std_logic;
	ackO_I: in std_logic);
  end component;
  -- read logic
  signal rd_rst, rd_almost_full, rd_full: std_logic;
  signal sack_rd, msel_rd, rd_stbO_O, mack_rd: std_logic;
  signal mrnw1, srnw_q: std_logic;-- internal master read not write
  -- write logic
  signal wr_almost_empty: std_logic;
  signal ssel_wr, sack_wr, msel_wr, mack_wr: std_logic;
  signal mbe_rd, mbe_wr, sbe_q: std_logic_vector(0 to dwidth/8 -1); -- master byte enable for read and write
  signal wr_din, wr_dout: std_logic_vector(0 to dwidth + dwidth/8 -1); -- write fifo data in and out
  -- address logic
  signal ma_q: std_logic_vector(0 to awidth-1); -- address counter output
  signal load_ma_q, ma_valid: std_logic; -- load address counter, -- valid ma_q value
  signal seqAddrQ, mseqAddr1: std_logic; -- sequential address (detect any low value on sseqAddr)
begin
  mseqAddr<= mseqAddr1; -- or mseqAddr_wr;  
  msel<= msel_wr or msel_rd;
  sack<= sack_rd or sack_wr;
  mbe<= mbe_rd when mrnw1='1' else mbe_wr;
  mrnw<= mrnw1;
----------------------------------------------------------------------
  -- read logic  
  -- no read logic
  grd0: if fifo_rd_size<0 generate
      sdo<= (others=>'0');
      mbe_rd<= (others=> '0');
      msel_rd<= '0';
      sack_rd<= '0';
   end generate;

  grd1: if fifo_rd_size>= 0 generate
    frd: fifo
     generic map(adr_width=> fifo_rd_size, data_width=> dwidth)
     port map(clk=> clk, arst=> arst, 
	   rst=> rd_rst, almost_full=> rd_almost_full, almost_empty=> open, full=> rd_full, half_full=> open,
	   -- data input signals (slave)
	   datI_I=> mdi, stbI_I=> mack_rd, ackI_O=> open, -- master to master connection so different handshaking
	   -- data output signals (master)
	   datO_O=> sdo, stbO_O=> rd_stbO_O, ackO_I=> sack_rd);
  
    sack_rd<= rd_stbO_O and ma_valid and ssel and srnw;
    mack_rd<= mack and mrnw1;
    msel_rd<= not rd_full and mrnw1 and ma_valid;
    rd_rst<= not ma_valid or (not sseqAddr and sack_rd);
    
   end generate;
    --------------------------------------------------------------------------------
   -- write logic
   g_wr0: if fifo_wr_size<0 generate -- do not use write logic
         mdo<= (others=> '0');
         mbe_wr<= (others=> '0');
         msel_wr<= '0';
         sack_wr<= '0';
		 ssel_wr<= '0';
		 wr_almost_empty<= '1'; -- not to generate seqAddr when rnw='0'
  end generate;           
  
  gwr1: if fifo_wr_size>=0 generate
    fwr: fifo
     generic map(adr_width=> fifo_wr_size, data_width=> dwidth + dwidth/8)
     port map(clk=> clk, arst=> arst, rst=> '0',
       almost_full=> open, almost_empty=> wr_almost_empty, full=> open, half_full=> open,
	   -- data input signals (slave)
	   datI_I=> wr_din, stbI_I=> ssel_wr, ackI_O=> sack_wr,
	   -- data output signals (master)
	   datO_O=> wr_dout, stbO_O=> msel_wr, ackO_I=> mack_wr);

     ssel_wr<= ssel and not srnw;
     mack_wr<= not mrnw1 and mack;
     wr_din<= sdi & sbe;
     mdo<= wr_dout(0 to dwidth-1);
     mbe_wr<= wr_dout(dwidth to dwidth + dwidth/8 -1);
  end generate;
  ---------------------------------------------------------------------------------------------
  -- address logic
  g_seqAddr1: if use_seqAddr>0 generate
    process(clk,arst) begin
      if arst='1' then seqAddrQ<= '0';
      elsif clk'event and clk='1' then
        if seqAddrQ='1' then 
          seqAddrQ<= sseqAddr; -- detect every low value on seqAddr
        elsif mseqAddr1='0' and ssel='1' then -- the low state on sseqAddr must be transmitted to the mseqAddr therefore mseqAddr1='0'
          seqAddrQ<= sseqAddr;
        end if;
      end if;
     end process;
  end generate;
  
  
  gseqAddr0: if use_seqAddr<=0 generate
      seqAddrQ<= '0';
  end generate;
  
  mseqAddr1<= '0' when use_seqAddr<=0 else
    sseqAddr when mrnw1='1' and fifo_rd_size=0 else
    seqAddrQ when mrnw1='1' and fifo_rd_size>0 else
    not wr_almost_empty or seqAddrQ; -- for write
  
  -- generate internal address counter
  g_maq: if fifo_rd_size>0 or fifo_wr_size>0 generate
   process(arst, clk) begin
    if arst='1' then ma_q<= (others=>'0'); ma_valid<= '0'; srnw_q<= '0';
    elsif clk'event and clk='1' then
      if load_ma_q= '1' then
        ma_q<= sa; -- latch the address from the slave
        ma_valid<= '1';
        srnw_q<= srnw; -- remember the cycle type
      else
        if (srnw_q='0' and ((wr_almost_empty='1' and mack='1' and seqAddrQ='0') or -- write buffer will be empty in the next clk cycle
            (msel_wr='0' and seqAddrQ= '0'))) or -- write buffer is empty and not seq Addressing
            ( srnw_q='1' and (seqAddrQ='0' or sseqAddr='0')) then -- the last cycle of the ack
              ma_valid<= '0';
        end if;          
        if mack='1' and use_seqAddr>0 then -- imcrease address by 1
           if dwidth=16 then 
             ma_q(awidth-1)<= '0'; ma_q(0 to awidth-2)<= ma_q(0 to awidth-2)+1;
           elsif dwidth= 32 then 
              ma_q(awidth-2 to awidth-1)<= "00"; ma_q(0 to awidth-3)<= ma_q(0 to awidth-3) + 1;
           end if;
        end if;
      end if;
    end if;
   end process;
   
   -- load ma_q counter
   load_ma_q<= '1' when ssel='1' and seqAddrQ='0' -- seqAddrQ is delayed by 1 clk so the address is latched in the first clk when seqAddrQ='0'
    -- and rd_stbO_O='0' and mack_rd='0'  -- read buffer is empty and will be empty (seqAddr='0')
    and (msel_wr='0' or (wr_almost_empty='1' and mack='1')) -- write fifo empty or will be empty after this clk
    else '0';
  end generate; -- use internal address counter
  
  -- do not generate address counter (direct input output connection)
  gmaq0: if fifo_rd_size<=0 and fifo_wr_size<=0 generate
	ma_valid<= '1';
  end generate;
  
  mrnw1<= srnw_q when fifo_wr_size>0  and fifo_rd_size>0 else -- both read and write uses address counter
  not msel_wr and srnw when fifo_rd_size=0 else -- direct read (no address counter for reading) and write using address counter
  srnw;
          
      
  -- address multiplexer
  ma<= sa when (mrnw1='1' and fifo_rd_size= 0) or -- read 
            (mrnw1='0' and fifo_wr_size=0) --write 
       else ma_q; -- use address from address counter  
          

-------------------------------------------------------------------------------
-- byte enable logic  
  gsbe_q: if (fifo_wr_size>0 or fifo_rd_size>0) and dwidth>8 generate
    process(arst, clk) begin
      if arst='1' then sbe_q<= (others=> '0');
      elsif clk'event and clk='1' then
        sbe_q<= sbe;
      end if;
    end process;
  end generate;
  
  gberd0: if fifo_rd_size=0 generate
      mbe_rd<= sbe;
  end generate;
  gberd8: if fifo_rd_size>0 and dwidth= 8 generate
      mbe_rd(0)<= msel_rd;
  end generate;
  gberd16: if fifo_rd_size>0 and dwidth= 16 generate
    mbe_rd<= sbe_q when use_seqAddr<=0 or seqAddrQ='0' else
        (not ma_q(awidth-1)) & '1';
  end generate;
  gberd32: if fifo_rd_size>0 and dwidth= 32 generate
    mbe_rd<= sbe_q when use_seqAddr<=0 or seqAddrQ='0' else
        (not ma_q(awidth-1) and not ma_q(awidth-1)) & -- MSB
        not ma_q(awidth-2) &
        not (ma_q(awidth-1) and ma_q(awidth-2)) &
        '1'; -- LSB
  end generate;
             
           
end arch;    
    
