-- opb2opb_dwidth module
-- this module can transfer data from one OPB Bus to another OPB bus 
-- with data width conversion

-- Author:  Ernest Jamro (AGH University of Science and Technology - Cracow, Poland) 
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement and author name is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset signal

entity opb2opb_dwidth is
  generic (c_opb_dwidth: integer:= 32; -- OPB data bus width (same for all buses - use c_dwidthM and c_dwidthS for real data widths
           c_opb_awidth: integer:= 32; -- OPB address bus width
		       c_sdwidth: integer:= 8; -- real OPB data width of the Slave side
           c_mdwidth: integer:= 16; -- real OPB data width of the Master size
           c_fifo_wr_awidth: integer:= 4; -- fifo address width:
             -- <0 - do not implement write logic at all
             -- 0 - direct connection (no fifo)
             -- 1 - only registers
             -- 4 fifo address width (2**c_fifo_wr_awidth location) - should be 4 (optimal for small distributed BRAMs)
		       c_fifo_rd_awidth: integer:= 0; -- fifo for reading - see c_fifo_wr_awidth for more information
           c_use_seqAddr: integer:= 0; -- use sequential addressing
             -- 0 - do not use seq Addr
             -- 1 - use seqAddr
           c_use_toutSup: integer:= 0; -- =0 - do not generate OPB_toutSup, =1- generate OPB_toutSup when this device is selected
           c_use_ByteEnable: integer:= 1;
           c_single_master: integer:= 0;
		       c_speedrate : integer:= 0; -- should be 0 unless the minimum period time condition is not met
           c_transfer_size: integer:= 4; -- max transfer size when bus is locked (valid only when seqAddr>0)
           C_MSBADDR   : std_logic_vector(0 to 31) := X"FFFF_FFFF"; -- MSBs of master address - used only when fifferent form FFFF_FFFF
           C_BASEADDR   : std_logic_vector(0 to 31) := X"0000_0000"; -- only single address region is decoded
           C_HIGHADDR   : std_logic_vector(0 to 31) := X"0000_01FF");
	port (OPB_clk: in STD_LOGIC; 
	    OPB_rst: in STD_LOGIC;
		-- OPB Slave interface
       SOPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
       SOPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
       SOPB_RNW     : in std_logic;
       SOPB_select  : in std_logic;
       SOPB_seqAddr : in std_logic;
       SOPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
       Sl_errAck  : out std_logic;
       Sl_retry   : out std_logic;
       Sl_toutSup : out std_logic;
       Sl_xferAck : out std_logic;
       -- OPB Master interface
	   M_ABus : out STD_LOGIC_VECTOR(0 TO c_opb_awidth-1); -- is already gated by M_Select
	   M_DBus : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
       M_BE : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth/8 -1); 
	   M_RNW : out STD_LOGIC;
	   M_select : out STD_LOGIC;
	   M_seqAddr : out STD_LOGIC;
	   M_busLock : out STD_LOGIC;
	   M_request : out STD_LOGIC;
	   MOPB_MGrant: in STD_LOGIC;
	   MOPB_pendReq: in STD_LOGIC;
	   MOPB_DBus : in STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
	   MOPB_errAck : in STD_LOGIC;
	   MOPB_retry : in STD_LOGIC;
	   MOPB_timeout: in STD_LOGIC;
	   MOPB_xferAck : in STD_LOGIC);
end opb2opb_dwidth;

architecture opb2opb_dwidth_arch of opb2opb_dwidth is
-- functions and components declarations
component ROC -- reset on Configuration
   port (O : out std_logic);
  end component;

component opb_master
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width, the LSBs of OPB_DBus='0'
    real_dwidth: integer:= 0; -- opb bus dwidth (=0 then the same as dwidth)
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal user address bus width
    transfer_size: integer:= 4; -- max transfer size when bus is locked (valid only when seqAddr>0)
    use_be: integer:= 1; -- >0 use Byte Enable Logic, <=0 assume full dwidth transfer all the time
    use_seqAddr: integer:= 1; -- =1 use sequential addressing, =0 -ignore seqAddr signal
    use_addr_counter: integer:= 0; -- <=0 - address is loaded with select signal, >0 use internal address counter, in this mode address (adr) is loaded when load_addr='1' (sel signal is ignored for address logic) and is internally increamented after every data transfer, for reading operation two options are available =1 start reading after load_addr='1', =2 start reading just after reset signal from address=0
    single_master: integer:= 0; -- =1 only a single master is atached to the OPB bus - seting this parameter allows for additional hardware savings, =0 more than one master is connected to the OPB bus
    fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
    fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
    wr_pipeline_latency: integer:= 0); -- user (external) write logic pipeline latency, (=0 - combinational logic) 1- only a single pipeline register - dwr are delayed by 1 clock in comparison to the address, ack- signal is ack for the address - data (di) are accepted after 1 clk latter (stb and ack signal is disregarded)
 port(clk, arst: std_logic; --clock and global asynchronous reset
		-- OPB Master interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO (opb_dwidth+7)/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    M_request : out STD_LOGIC;
      OPB_rst: in std_logic:= '0';
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
      -- internal simplified bus
      load_addr: in std_logic:= '0'; -- load internal address counter - signal is ignored when use_addr_counter<=0
      fifo_full, fifo_empty: out std_logic; -- fifo is full and empty
      fifo_almost_full, fifo_almost_empty: out std_logic; -- only one data can be transferred and fifo will be full or empty
      retry: out std_logic; -- FIFO state indicates that it is better to dwaw retry signal (this signal might be used by an opb_slave module connected to this opb_master module)
      di: in std_logic_vector(0 to dwidth-1):= (others=>'0'); -- data input
      do: out std_logic_vector(0 to dwidth-1); -- data output
      be: in std_logic_vector(0 to (dwidth+7)/8 -1):= (others=>'1'); -- byte enable
      a: in std_logic_vector(0 to awidth-1); -- address bus
      msb_adr: in std_logic_vector(0 to opb_awidth-1):=(others=>'0'); -- the MSB of the OPB address are specified by the msb_adr
      rnw: in std_logic; -- read not write
	    sel: in std_logic; -- OPB select signal
      seqAddr: in std_logic; -- seqAddressing - a (address) is incremented - block transfer - similar logic assert OPB_seqAddr
	    ack: out std_logic);
end component;

component OPB_slave
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_REAL_DWIDTH: integer:= 0; -- real OPB bus data width used during data transfer (=0 - same as dwidth)
    DWIDTH : integer:= 8; -- user width of the data - the rest LSBs are always 0s
    AWIDTH : integer:= 8; -- user address width
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	  C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
	   -- and the lower occupied area however the lower clock frequency is required. 
     -- Therefore to increase the clock frequency this parameter should be increased
	  -- user interface data and address width
    fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
    fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
    use_toutSup: integer:= 0; -- >0 generate timeout Suppress signal whenever this slave is selected
    use_be: integer:= 1; -- use Byte Enable logic;  <=0 - Byte enable are always '1' (only full data transfres are used)
    USE_SeqAddr : integer:= 1 -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
    );
  port ( clk : in std_logic:= '0'; -- clock signal
    arst: in std_logic:= '0'; -- asynchronous reset (needed only for simulation)
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_rst     : in std_logic:= '0';
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- user signals
	  sel: out std_logic; -- select - active when OPB_select and proper address
	  ack: in std_logic; -- acknowlenge - slave is ready for data transfer
    retry: in std_logic:= '0'; -- generate retry signal
	  rnw: out std_logic; -- read not write - same assert OPB_read not write
	  adr: out std_logic_vector(0 to awidth-1); -- address bus
	  drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	  dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	  be: out std_logic_vector(0 to (dwidth+7)/8 -1); -- byte enable 
	  seqAddr: out std_logic -- same as OPB_seqAddr
    );
end component;
function log2(arg: integer) return integer is
  variable ret: integer:= -1;
  begin
   for i in 0 to 32 loop
	  if 2**i >= arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
end log2;
function FindMsbAdr(ignore: integer) return std_logic_vector is
  constant ones: std_logic_vector(0 to c_opb_awidth-1):= (others=>'1');
 begin
  if C_MSBADDR = ones then
    return C_BASEADDR;
  else
    return C_MSBADDR;
  end if;
end FindMsbAdr;    
   

  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant awidth: integer:= log2(adr_space); -- slave address width
  constant msb_adr: std_logic_vector(0 to c_opb_awidth-1):= FindMsbAdr(0);
  signal adr: std_logic_vector(0 to awidth-1);
  signal dwr, drd: std_logic_vector(0 to c_sdwidth-1);
  signal be: std_logic_vector(0 to (c_sdwidth+7)/8-1);
  signal arst, sel, ack, rnw, retry: std_logic;

begin  
 sl: OPB_slave
  generic map( C_OPB_AWIDTH=> C_OPB_AWIDTH, C_OPB_DWIDTH=> C_OPB_DWIDTH,
    C_REAL_DWIDTH=> c_sdwidth, DWIDTH=> c_sdwidth, AWIDTH=> AWIDTH,
    C_BASEADDR=> C_BASEADDR, C_HIGHADDR=> C_HIGHADDR, C_SPEEDRATE=> C_SPEEDRATE,
    fifo_wr_awidth=> c_fifo_wr_awidth, fifo_rd_awidth=> c_fifo_rd_awidth,
    use_toutSup=> c_use_toutSup, use_be=> c_use_byteenable, USE_SeqAddr=> 0)
  port map( clk=> OPB_clk, arst=> arst, OPB_rst=> OPB_rst,
    -- OPB signals
    OPB_ABus=> SOPB_ABus, OPB_BE=> SOPB_BE, OPB_RNW=> SOPB_RNW,
    OPB_select=> SOPB_select, OPB_seqAddr=> SOPB_seqAddr, OPB_DBus=> SOPB_DBus, 
    Sl_DBus=> Sl_DBus, Sl_errAck=> Sl_errAck, Sl_retry=> Sl_retry,
    Sl_toutSup=> Sl_toutSup, Sl_xferAck=> Sl_xferAck,
	  -- user signals
	  sel=> sel, ack=> ack, rnw=> rnw, retry=> retry,
	  adr=> adr, drd=> drd, dwr=> dwr,
	  be=> be, seqAddr=> open);
    
    
 master: opb_master
  generic map(opb_dwidth=> c_opb_dwidth, dwidth=> c_sdwidth, real_dwidth=> c_mdwidth,
    opb_awidth=> c_opb_awidth, awidth=> awidth, transfer_size=> c_transfer_size,
    use_be=> c_use_byteenable, use_seqAddr=> c_use_seqAddr, use_addr_counter=> 0,
    single_master=> c_single_master, fifo_wr_awidth=> c_fifo_wr_awidth,
    fifo_rd_awidth=> c_fifo_rd_awidth, wr_pipeline_latency=> 0)
   port map (clk=> OPB_clk, arst=> arst,
		-- OPB Master interface
	    M_ABus=> M_ABus, M_BE=> M_BE, M_RNW=> M_RNW,
	    M_select=> M_select, M_seqAddr=> M_seqAddr, M_busLock=> M_busLock,
	    M_DBus=> M_DBus, M_request=> M_request,
      OPB_rst=> OPB_rst, OPB_MGrant=> MOPB_MGrant,
	    OPB_pendReq=> MOPB_pendReq, OPB_DBus=> MOPB_DBus,
	    OPB_errAck=> MOPB_errAck, OPB_retry=> MOPB_retry,
	    OPB_timeout=> MOPB_timeout, OPB_xferAck=> MOPB_xferAck,
      -- internal simplified bus
      load_addr=> '0', fifo_full=> open, fifo_empty=> open,
      fifo_almost_full=> open, fifo_almost_empty=> open, retry=> retry,
      di=> dwr, do=> drd, be=> be,
      a=> adr, msb_adr=> msb_adr, rnw=> rnw, sel=> sel,
      seqAddr=> SOPB_seqAddr, ack=> ack);
    
 -- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
 global_reset: ROC
	port map (O=> arst);

end opb2opb_dwidth_arch;